@@ -1,4 +1,4 @@
-#include <stdio.h>
+int puts(const char *s);
 
 int main(int argc, char *argv[])
 {
